from datetime import datetime
from dateutil import parser
from .standing import Standing
from .match import Match
from typing import List, Dict, Any, Optional


class Bracket:
    type: str
    series_style: str
    style: str
    teams_count: int
    has_third_place_match: bool
    rounds_count: int

    def __init__(self, data: dict) -> None:
        self.type = data.get("type")
        self.series_style = data.get("seriesStyle")
        self.style = data.get("style")
        self.teams_count = data.get("teamsCount")
        self.has_third_place_match = data.get("hasThirdPlaceMatch")
        self.rounds_count = data.get("roundsCount")


class Stage:
    id: str
    name: str
    start_time: Optional[datetime]
    has_match_checkin: bool
    has_checkin_timer: bool
    has_confirm_score: bool
    bracket: Bracket
    created_at: Optional[datetime]
    updated_at: Optional[datetime]
    has_started: bool
    started_at: Optional[datetime]

    standings: List[Standing]
    matches: List[Match]

    raw_matches: List[dict]
    raw_data = Dict[str, Any]

    def __init__(self, data: dict) -> None:
        self.id = data.get("_id")
        self.name = data.get("name")
        if "startTime" in data:
            self.start_time = parser.isoparse(data.get("startTime"))
        self.has_match_checkin = data.get("hasMatchCheckin")
        self.has_checkin_timer = data.get("hasCheckinTimer")
        self.has_confirm_score = data.get("hasConfirmScore")
        self.bracket = Bracket(data.get("bracket"))
        if "createdAt" in data:
            self.created_at = parser.isoparse(data.get("createdAt"))
        if "updatedAt" in data:
            self.updated_at = parser.isoparse(data.get("updatedAt"))
        self.has_started = data.get("hasStarted")
        if "startedAt" in data:
            self.started_at = parser.isoparse(data.get("startedAt"))
        self.standings = []
        self.matches = []
        self.raw_matches = []
        self.raw_data = data

    def add_standing(self, standing: Standing):
        self.standings.append(standing)

    def add_match(self, match: Match):
        self.matches.append(match)




import asyncio
import aiohttp
from typing import Optional, List, Union

from .models import Tournament, Stage, Match, SwissStanding, LadderStanding, BracketStanding, RoundRobinStanding


class TournamentClient:
    def __init__(self, headers: dict = None):
        self.headers = headers or {}

    @staticmethod
    def get_bracket_rankings(no_teams: int, is_double_elim: bool, third_place: Union[bool, None] = None) -> List[int]:
        """
        Get the rankings for a bracket
        :param third_place: If there's a third place match (for single elimination)
        :param is_double_elim: if bracket is double elimination
        :param no_teams: number of teams in bracket
        :return: the ranking boundaries for the bracket
        """
        rank = (2 ** (no_teams - 1).bit_length())
        return_list = []
        if not is_double_elim:
            if third_place:
                return_list = return_list + [4, 3, 2, 1]
            else:
                return_list = return_list + [2, 1]
            while rank > return_list[0]:
                rank = int(rank // 2)
                return_list.append(rank + 1)
        else:
            return_list = return_list + [4, 3, 2, 1]
            subtractor = int(rank // 4)
            while subtractor >= 2:
                rank = rank - subtractor
                return_list.append(rank + 1)
                rank = rank - subtractor
                return_list.append(rank + 1)
                subtractor = int(subtractor // 2)
        return_list.sort(reverse=True)
        return return_list

    async def __fallback_get_elimination_standing(self, stage_data: Stage, tournament_data: Tournament) -> Stage:
        team_standings = {}
        placement_data = {}
        for team in tournament_data.teams:
            team_standings[team.persistent_team_id] = BracketStanding({}, team)
        for match in stage_data.raw_matches:
            try:
                if ("team" in match.get("top", {})) and "team" in match.get("bottom", {}):
                    top_winner = match.get("top", {}).get("winner", False)
                    bottom_winner = match.get("bottom", {}).get("winner", False)
                    if top_winner:
                        team_standings[match["bottom"]["team"]["persistentTeamID"]].set_losses += 1
                        team_standings[match["top"]["team"]["persistentTeamID"]].set_wins += 1
                    elif bottom_winner:
                        team_standings[match["top"]["team"]["persistentTeamID"]].set_losses += 1
                        team_standings[match["bottom"]["team"]["persistentTeamID"]].set_wins += 1
                    else:
                        team_standings[match["top"]["team"]["persistentTeamID"]].set_ties += 1
                        team_standings[match["bottom"]["team"]["persistentTeamID"]].set_ties += 1
            except Exception:
                continue
        if stage_data.bracket.style == "double":
            # Double elimination
            rankings = self.get_bracket_rankings(stage_data.bracket.teams_count, True)
            for r in range(len(rankings)):
                placement_data[r + 1] = []
            for match in stage_data.raw_matches:
                try:
                    if match.get("inConsolationBracket", False):
                        if match["top"]["winner"]:
                            placement_data[match["roundNumber"]].append(match["bottom"]["team"]["persistentTeamID"])
                        elif match["bottom"]["winner"]:
                            placement_data[match["roundNumber"]].append(match["top"]["team"]["persistentTeamID"])
                        else:
                            placement_data[match["roundNumber"]].append(match["top"]["team"]["persistentTeamID"])
                            placement_data[match["roundNumber"]].append(match["bottom"]["team"]["persistentTeamID"])
                except Exception:
                    continue
            match = stage_data.raw_matches[-1]
            if match.get('top', {}).get('team', {}) and match.get('bottom', {}).get('team', {}):
                top_index = len(rankings)
                if match["top"]["winner"]:
                    placement_data[top_index].append(match["top"]["team"]["persistentTeamID"])
                    placement_data[top_index - 1].append(match["bottom"]["team"]["persistentTeamID"])
                elif match["bottom"]["winner"]:
                    placement_data[top_index].append(match["bottom"]["team"]["persistentTeamID"])
                    placement_data[top_index - 1].append(match["top"]["team"]["persistentTeamID"])
                else:
                    placement_data[top_index - 1].append(match["top"]["team"]["persistentTeamID"])
                    placement_data[top_index - 1].append(match["bottom"]["team"]["persistentTeamID"])
        elif stage_data.bracket.style == "single":
            rankings = self.get_bracket_rankings(stage_data.bracket.teams_count, False,
                                                 stage_data.bracket.has_third_place_match)
            for r in range(len(rankings)):
                placement_data[r + 1] = []
            for match in stage_data.raw_matches:
                try:
                    if ("team" in match.get("top", {})) and "team" in match.get("bottom", {}):
                        if not match.get("matchType", "") == "loser":
                            if match.get("roundNumber", 0) == stage_data.bracket.rounds_count:
                                first = len(rankings)
                                second = len(rankings) - 1
                                if match["top"]["winner"]:
                                    placement_data[first].append(match["top"]["team"]["persistentTeamID"])
                                    placement_data[second].append(match["bottom"]["team"]["persistentTeamID"])
                                elif match["bottom"]["winner"]:
                                    placement_data[second].append(match["top"]["team"]["persistentTeamID"])
                                    placement_data[first].append(match["bottom"]["team"]["persistentTeamID"])
                                else:
                                    placement_data[second].append(match["top"]["team"]["persistentTeamID"])
                                    placement_data[second].append(match["bottom"]["team"]["persistentTeamID"])
                            else:
                                # Skip round 2 if we have a third place match
                                if stage_data.bracket.has_third_place_match:
                                    if match.get("roundNumber", 0) == stage_data.bracket.rounds_count - 1:
                                        continue
                                placement_number = match.get("roundNumber", 0)
                                if match["top"]["winner"]:
                                    placement_data[placement_number].append(match["bottom"]["team"]["persistentTeamID"])
                                elif match["bottom"]["winner"]:
                                    placement_data[placement_number].append(match["top"]["team"]["persistentTeamID"])
                                else:
                                    placement_data[placement_number].append(match["top"]["team"]["persistentTeamID"])
                                    placement_data[placement_number].append(match["bottom"]["team"]["persistentTeamID"])
                        else:  # In a 3rd place match
                            third = len(rankings)-2
                            fourth = len(rankings)-3
                            if match["top"]["winner"]:
                                placement_data[third].append(match["top"]["team"]["persistentTeamID"])
                                placement_data[fourth].append(match["bottom"]["team"]["persistentTeamID"])
                            elif match["bottom"]["winner"]:
                                placement_data[third].append(match["bottom"]["team"]["persistentTeamID"])
                                placement_data[fourth].append(match["top"]["team"]["persistentTeamID"])
                            else:
                                placement_data[fourth].append(match["top"]["team"]["persistentTeamID"])
                                placement_data[fourth].append(match["bottom"]["team"]["persistentTeamID"])
                except Exception:
                    continue
        else:
            raise ValueError(f"Invalid bracket style {stage_data.bracket.style}")
        for key, value in placement_data.items():
            set_value = list(set(value))  # Remove duplicate teams in a round
            for team_id in set_value:
                team_standings[team_id].set_place(rankings[key - 1])
                stage_data.add_standing(team_standings[team_id])
        return stage_data

    async def __get_elimination_standing(self, stage_data: Stage, tournament_data: Tournament) -> Stage:
        async with aiohttp.ClientSession() as session:
            async with session.get(
                    f"https://dtmwra1jsgyb0.cloudfront.net/stages/{stage_data.id}/standings",
                    headers=self.headers) as resp:
                if resp.status != 200:
                    raise ValueError(f"Invalid ID {stage_data.id}")
                standing_data = await resp.json()
                if (not standing_data) or ("wins" not in standing_data[0]):
                    return await self.__fallback_get_elimination_standing(stage_data, tournament_data)
                for standing in standing_data:
                    stage_data.add_standing(
                        BracketStanding(standing, tournament_data.get_team_from_team_id(standing["teamID"])))
        return stage_data

    async def __get_ladder_standing(self, stage_data: Stage, tournament_data: Tournament) -> Stage:
        # Get placements data
        placement = {}
        async with aiohttp.ClientSession() as session:
            async with session.get(
                    f"https://dtmwra1jsgyb0.cloudfront.net/stages/{stage_data.id}/ladder-standings",
                    headers=self.headers) as resp:
                if resp.status != 200:
                    raise ValueError(f"Invalid ID {stage_data.id}")
                initial_data = await resp.json()
                pages = initial_data["total"] // 10 + 1
                for x in range(len(initial_data["standings"])):
                    placement[initial_data["standings"][x]["teamID"]] = x + 1
                for page in range(1, pages + 1):
                    skip = page * 10
                    async with session.get(
                            f"https://dtmwra1jsgyb0.cloudfront.net/stages/{stage_data.id}/ladder-standings?filter=all&skip={skip}&limit=10",
                            headers=self.headers) as resp:
                        if resp.status != 200:
                            raise ValueError(f"Invalid ID {stage_data.id}")
                        data = await resp.json()
                        for x in range(len(data["standings"])):
                            placement[data["standings"][x]["teamID"]] = x + 1 + skip
            # Get standing data
            async with session.get(
                    f"https://dtmwra1jsgyb0.cloudfront.net/stages/{stage_data.id}/latest-round-standings",
                    headers=self.headers) as resp:
                if resp.status != 200:
                    raise ValueError(f"Invalid ID {stage_data.id}")
                standing_data = await resp.json()
                for standing in standing_data:
                    stage_data.add_standing(
                        LadderStanding(standing, tournament_data.get_team_from_team_id(standing["teamID"]),
                                       placement.get(standing["teamID"], None)))
        return stage_data

    async def __get_swiss_standings(self, stage_data: Stage, tournament_data: Tournament) -> Stage:
        async with aiohttp.ClientSession() as session:
            async with session.get(
                    f"https://dtmwra1jsgyb0.cloudfront.net/stages/{stage_data.id}/latest-round-standings",
                    headers=self.headers) as resp:
                if resp.status != 200:
                    raise ValueError(f"Invalid ID {stage_data.id}")
                standing_data = await resp.json()
                # Sort placement for completed and DNF
                completed = []
                dnf = []
                for standing in standing_data:
                    if standing["disqualified"]:
                        dnf.append(standing)
                    else:
                        completed.append(standing)
                standing_data = completed + dnf
                for n in range(len(standing_data)):
                    standing = standing_data[n]
                    if "teamID" in standing:
                        stage_data.add_standing(
                            SwissStanding(standing, tournament_data.get_team_from_team_id(standing["teamID"]), n + 1))
        return stage_data

    async def __get_round_robin_standings(self, stage_data: Stage, tournament_data: Tournament) -> Stage:
        async with aiohttp.ClientSession() as session:
            team_standings: dict = {}
            # Load standings data into players
            async with session.get(
                    f"https://dtmwra1jsgyb0.cloudfront.net/stages/{stage_data.id}/standings",
                    headers=self.headers) as resp:
                if resp.status != 200:
                    raise ValueError(f"Invalid ID {stage_data.id}")
                standing_data = await resp.json()
                for standing in standing_data:
                    standing = RoundRobinStanding(standing, tournament_data.get_team_from_team_id(standing["teamID"]))
                    stage_data.add_standing(standing)
                    team_standings[standing.team.id] = standing
            # Get RR groups
            groups: List[List[RoundRobinStanding]] = []
            async with session.get(
                    f"https://api.battlefy.com/stages/{stage_data.id}?extend%5Bgroups%5D%5Bteams%5D=true",
                    headers=self.headers) as resp:
                if resp.status != 200:
                    raise ValueError(f"Invalid ID {stage_data.id}")
                group_request = await resp.json()
                for group in group_request[0]["groups"]:
                    working_group = []
                    for team in group["teamIDs"]:
                        working_group.append(team_standings[team])
                    groups.append(working_group)
        #  Calculate RR standings
        for rr_group in groups:
            sorted_group = sorted(rr_group, key=lambda x: (x.points, x.set_wins, x.match_win_amongst_tied,
                                                           x.game_win_lose_difference, x.total_win_lose_difference),
                                  reverse=True)
            for n in range(len(sorted_group)):
                sorted_group[n].place = n + 1
                if n > 0:
                    if sorted_group[n].set_wins == sorted_group[n - 1].set_wins and \
                            sorted_group[n].match_win_amongst_tied == sorted_group[n - 1].match_win_amongst_tied and \
                            sorted_group[n].game_win_lose_difference == sorted_group[n - 1].game_win_lose_difference and \
                            sorted_group[n].total_win_lose_difference == sorted_group[n - 1].total_win_lose_difference:
                        sorted_group[n].place = sorted_group[n - 1].place
        return stage_data

    async def __get_standings(self, stage_data: Stage, tournament_data: Tournament) -> Stage:
        if stage_data.bracket.type == "elimination":
            return await self.__get_elimination_standing(stage_data, tournament_data)
        elif stage_data.bracket.type == "swiss":
            return await self.__get_swiss_standings(stage_data, tournament_data)
        elif stage_data.bracket.type == "ladder":
            return await self.__get_ladder_standing(stage_data, tournament_data)
        elif stage_data.bracket.type == "roundrobin":
            return await self.__get_round_robin_standings(stage_data, tournament_data)
        else:
            raise ValueError(f"Invalid stage type {stage_data.bracket.type}")

    async def __get_additional_data(self, stage_data: Stage, tournament_data: Tournament) -> Stage:
        async with aiohttp.ClientSession() as session:
            async with session.get(
                    f"https://api.battlefy.com/stages/{stage_data.id}?extend[matches][top.team][players][user]=true&extend[matches][top.team][persistentTeam]=true&extend[matches][bottom.team][players][user]=true&extend[matches][bottom.team][persistentTeam]=true&extend[groups][teams]=true&extend[groups][matches][top.team][players][user]=true&extend[groups][matches][top.team][persistentTeam]=true&extend[groups][matches][bottom.team][players][user]=true&extend[groups][matches][bottom.team][persistentTeam]=true",
                    headers=self.headers) as resp:
                if resp.status != 200:
                    raise ValueError(f"Invalid ID {stage_data.id}")
                stage_request = await resp.json()
                stage_request = stage_request[0]
                for match in stage_request.get("matches", []):
                    for side in ["top", "bottom"]:
                        if side in match:
                            if "team" in match[side]:
                                for player in match[side]["team"]["players"]:
                                    if ("user" in player) and ("persistentPlayerID" in player):
                                        if ("accounts" in player["user"]) and ("discord" in player["user"]["accounts"]):
                                            player_entry = tournament_data.get_player_from_persistent_player_id(
                                                player.get("persistentPlayerID", None)
                                            )
                                            if player_entry:
                                                player_entry.discord_id = player["user"]["accounts"]["discord"].get(
                                                    "accountID",
                                                    "")
        return stage_data

    async def __get_stage(self, stage_data: Stage, tournament_data: Tournament) -> Stage:
        async with aiohttp.ClientSession() as session:
            async with session.get(
                    f"https://api.battlefy.com/stages/{stage_data.id}?extend[matches][top.team][players][user]=true&extend[matches][top.team][persistentTeam]=true&extend[matches][bottom.team][players][user]=true&extend[matches][bottom.team][persistentTeam]=true&extend[groups][teams]=true&extend[groups][matches][top.team][players][user]=true&extend[groups][matches][top.team][persistentTeam]=true&extend[groups][matches][bottom.team][players][user]=true&extend[groups][matches][bottom.team][persistentTeam]=true",
                    headers=self.headers) as resp:
                if resp.status != 200:
                    raise ValueError(f"Invalid ID {stage_data.id}")
                stage_request = await resp.json()
                stage_request = stage_request[0]
                stage_data.raw_matches = stage_request.get("matches", [])
                for match in stage_data.raw_matches:
                    top_team_id = match.get("top", {}).get("team", {}).get("persistentTeamID", None)
                    bottom_team_id = match.get("bottom", {}).get("team", {}).get("persistentTeamID", None)
                    if top_team_id and bottom_team_id:
                        top_team = tournament_data.get_team_from_persistent_team_id(top_team_id)
                        bottom_team = tournament_data.get_team_from_persistent_team_id(bottom_team_id)
                        if top_team and bottom_team:
                            stage_data.add_match(Match(match_data=match, top=top_team, bottom=bottom_team))
        await self.__get_standings(stage_data, tournament_data)
        return stage_data

    async def get_tournament(self, tournament_id: str) -> Optional[Tournament]:
        async with aiohttp.ClientSession() as session:
            async with session.get(
                    f"https://api.battlefy.com/tournaments/{tournament_id}?extend%5Bstages%5D%5Bmatches%5D=true&extend%5Bstages.matches.top.team.persistentTeam%5D=1&extend%5Bstages.matches.bottom.team.persistentTeam%5D=1&extend%5Bstages.matches.top.team.players%5D=1&extend%5Bstages.matches.bottom.team.players%5D=1&extend%5Bstages.matches.top.team%5D=1&extend%5Bstages.matches.bottom.team%5D=1",
                    headers=self.headers
            ) as resp:
                if resp.status != 200:
                    raise ValueError("Invalid ID")
                tournament_data = await resp.json()
                tournament = Tournament(data=tournament_data[0])
            async with session.get(f"https://dtmwra1jsgyb0.cloudfront.net/tournaments/{tournament_id}/teams") as resp:
                if resp.status != 200:
                    raise ValueError("Invalid ID")
                team_data = await resp.json()
                tournament.add_teams(team_data)
        async with asyncio.TaskGroup() as g:
            for stage in tournament.stages:
                g.create_task(self.__get_stage(stage, tournament))
        async with asyncio.TaskGroup() as g:
            for stage in tournament.stages:
                g.create_task(self.__get_additional_data(stage, tournament))
        return tournament



from dataclasses import dataclass, InitVar, field
from .team import Team
from datetime import datetime
from dateutil import parser


@dataclass
class Game:
    game_data: InitVar[dict]
    game_id: str = field(init=False)
    winner: int = field(init=False, default=-1)  # Top = 0, Bottom = 1
    created_at: datetime = field(init=False, default=None)

    def __post_init__(self, game_data: dict):
        if created_at := game_data.get("createdAt"):
            self.created_at = parser.isoparse(created_at)
        self.game_id = game_data.get("_id", None)
        stats = game_data.get("stats")
        if stats:
            if stats["top"]["winner"]:
                self.winner = 0
            elif stats["bottom"]["winner"]:
                self.winner = 1


@dataclass
class Match:
    match_data: InitVar[dict]
    # attributes
    top: Team
    bottom: Team
    id: str = field(init=False, default="")
    games: list = field(init=False, default_factory=list)
    created_at: datetime = field(init=False, default=None)
    top_win: int = field(init=False, default=0)
    bottom_win: int = field(init=False, default=0)

    def __post_init__(self, match_data: dict):
        self.id = match_data.get("_id")
        if created_at := match_data.get("createdAt"):
            self.created_at = parser.isoparse(created_at)
        if "stats" in match_data:
            for g in match_data.get("stats", []):
                self.games.append(Game(g))
            for game in self.games:
                if game.winner == 0:
                    self.top_win += 1
                elif game.winner == 1:
                    self.bottom_win += 1
        else:
            self.top_win = match_data.get("top", {}).get("score", 0)
            self.bottom_win = match_data.get("bottom", {}).get("score", 0)


    @property
    def valid_games(self):
        valid = Team
        for g in self.games:
            if g.winner == -1:
                valid = field()
        return valid



from enum import Enum
from dataclasses import dataclass
from typing import Optional, Any, List, TypeVar, Type, cast, Callable
from datetime import datetime
import dateutil.parser


T = TypeVar("T")
EnumT = TypeVar("EnumT", bound=Enum)


def from_none(x: Any) -> Any:
    assert x is None
    return x


def from_union(fs, x):
    for f in fs:
        try:
            return f(x)
        except:
            pass
    assert False


def from_str(x: Any) -> str:
    assert isinstance(x, str)
    return x


def to_enum(c: Type[EnumT], x: Any) -> EnumT:
    assert isinstance(x, c)
    return x.value


def from_datetime(x: Any) -> datetime:
    return dateutil.parser.isoparse(x)


def from_int(x: Any) -> int:
    assert isinstance(x, int) and not isinstance(x, bool)
    return x


def from_bool(x: Any) -> bool:
    assert isinstance(x, bool)
    return x


def to_class(c: Type[T], x: Any) -> dict:
    assert isinstance(x, c)
    return cast(Any, x).to_dict()


def from_list(f: Callable[[Any], T], x: Any) -> List[T]:
    assert isinstance(x, list)
    return [f(y) for y in x]


@dataclass
class Game:
    id: Optional[str] = None
    name: Optional[str] = None
    icon_url: Optional[str] = None
    image_url: Optional[str] = None
    background_url: Optional[str] = None
    slug: Optional[str] = None

    @staticmethod
    def from_dict(obj: Any) -> 'Game':
        assert isinstance(obj, dict)
        id = from_union([from_str, from_none], obj.get("_id"))
        name = from_union([from_str, from_none], obj.get("name"))
        icon_url = from_union([from_str, from_none], obj.get("iconUrl"))
        image_url = from_union([from_str, from_none], obj.get("imageUrl"))
        background_url = from_union([from_str, from_none], obj.get("backgroundUrl"))
        slug = from_union([from_str, from_none], obj.get("slug"))
        return Game(id, name, icon_url, image_url, background_url, slug)

    def to_dict(self) -> dict:
        result: dict = {}
        result["_id"] = from_union([from_str, from_none], self.id)
        result["name"] = from_union([from_str, from_none], self.name)
        result["iconUrl"] = from_union([from_str, from_none], self.icon_url)
        result["imageUrl"] = from_union([from_str, from_none], self.image_url)
        result["backgroundUrl"] = from_union([from_str, from_none], self.background_url)
        result["slug"] = from_union([from_str, from_none], self.slug)
        return result


class Status(Enum):
    REGISTRATION_CLOSED = "registration-closed"


class TypeEnum(Enum):
    TEAM = "team"


@dataclass
class Tournament:
    id: Optional[str] = None
    start_time: Optional[datetime] = None
    players_per_team: Optional[int] = None
    name: Optional[str] = None
    banner_url: Optional[str] = None
    is_published: Optional[bool] = None
    is_public: Optional[bool] = None
    registration_enabled: Optional[bool] = None
    slug: Optional[str] = None
    game_id: Optional[str] = None
    game: Optional[Game] = None
    teams_count: Optional[int] = None
    status: Optional[Status] = None

    @staticmethod
    def from_dict(obj: Any) -> 'Tournament':
        assert isinstance(obj, dict)
        id = from_union([from_str, from_none], obj.get("_id"))
        start_time = from_union([from_datetime, from_none], obj.get("startTime"))
        players_per_team = from_union([from_int, from_none], obj.get("playersPerTeam"))
        name = from_union([from_str, from_none], obj.get("name"))
        banner_url = from_union([from_str, from_none], obj.get("bannerUrl"))
        is_published = from_union([from_bool, from_none], obj.get("isPublished"))
        is_public = from_union([from_bool, from_none], obj.get("isPublic"))
        registration_enabled = from_union([from_bool, from_none], obj.get("registrationEnabled"))
        slug = from_union([from_str, from_none], obj.get("slug"))
        game_id = from_union([from_str, from_none], obj.get("gameID"))
        game = from_union([Game.from_dict, from_none], obj.get("game"))
        teams_count = from_union([from_int, from_none], obj.get("teamsCount"))
        status = from_union([Status, from_none], obj.get("status"))
        return Tournament(id, start_time, players_per_team, name, banner_url, is_published, is_public, registration_enabled, slug, game_id, game, teams_count, status)

    def to_dict(self) -> dict:
        result: dict = {}
        result["_id"] = from_union([from_str, from_none], self.id)
        result["startTime"] = from_union([lambda x: x.isoformat(), from_none], self.start_time)
        result["playersPerTeam"] = from_union([from_int, from_none], self.players_per_team)
        result["name"] = from_union([from_str, from_none], self.name)
        result["bannerUrl"] = from_union([from_str, from_none], self.banner_url)
        result["isPublished"] = from_union([from_bool, from_none], self.is_published)
        result["isPublic"] = from_union([from_bool, from_none], self.is_public)
        result["registrationEnabled"] = from_union([from_bool, from_none], self.registration_enabled)
        result["slug"] = from_union([from_str, from_none], self.slug)
        result["gameID"] = from_union([from_str, from_none], self.game_id)
        result["game"] = from_union([lambda x: to_class(Game, x), from_none], self.game)
        result["teamsCount"] = from_union([from_int, from_none], self.teams_count)
        result["status"] = from_union([lambda x: to_enum(Status, x), from_none], self.status)
        return result


@dataclass
class OrgSearch:
    total: int
    tournaments: List[Tournament]

    @staticmethod
    def from_dict(obj: Any) -> 'Welcome':
        assert isinstance(obj, dict)
        total = from_int(obj.get("total"))
        tournaments = from_list(Tournament.from_dict, obj.get("tournaments"))
        return OrgSearch(total, tournaments)

    def to_dict(self) -> dict:
        result: dict = {}
        result["total"] = from_int(self.total)
        result["tournaments"] = from_list(lambda x: to_class(Tournament, x), self.tournaments)
        return result


def org_search_from_dict(s: Any) -> OrgSearch:
    return OrgSearch.from_dict(s)


def tournaments_from_org_search(tournaments: List):
    return from_list(Tournament.from_dict, tournaments)



from dataclasses import dataclass
from typing import Any, Optional, List, TypeVar, Type, cast, Callable
from enum import Enum
from datetime import datetime
import dateutil.parser


T = TypeVar("T")
EnumT = TypeVar("EnumT", bound=Enum)


def from_str(x: Any) -> str:
    assert isinstance(x, str)
    return x


def from_none(x: Any) -> Any:
    assert x is None
    return x


def from_union(fs, x):
    for f in fs:
        try:
            return f(x)
        except:
            pass
    assert False


def from_bool(x: Any) -> bool:
    assert isinstance(x, bool)
    return x


def from_datetime(x: Any) -> datetime:
    return dateutil.parser.isoparse(x)


def from_int(x: Any) -> int:
    assert isinstance(x, int) and not isinstance(x, bool)
    return x


def to_enum(c: Type[EnumT], x: Any) -> EnumT:
    assert isinstance(x, c)
    return x.value


def to_class(c: Type[T], x: Any) -> dict:
    assert isinstance(x, c)
    return cast(Any, x).to_dict()


def is_type(t: Type[T], x: Any) -> T:
    assert isinstance(x, t)
    return x


def from_list(f: Callable[[Any], T], x: Any) -> List[T]:
    assert isinstance(x, list)
    return [f(y) for y in x]


def from_float(x: Any) -> float:
    assert isinstance(x, (float, int)) and not isinstance(x, bool)
    return float(x)


def to_float(x: Any) -> float:
    assert isinstance(x, float)
    return x


@dataclass
class AdminEmailSubscriptions:
    pass

    @staticmethod
    def from_dict(obj: Any) -> 'AdminEmailSubscriptions':
        assert isinstance(obj, dict)
        return AdminEmailSubscriptions()

    def to_dict(self) -> dict:
        result: dict = {}
        return result


@dataclass
class AdminEquipment:
    face: Optional[str] = None

    @staticmethod
    def from_dict(obj: Any) -> 'AdminEquipment':
        assert isinstance(obj, dict)
        face = from_union([from_str, from_none], obj.get("face"))
        return AdminEquipment(face)

    def to_dict(self) -> dict:
        result: dict = {}
        result["face"] = from_union([from_str, from_none], self.face)
        return result


@dataclass
class AdminFeatures:
    onboarded_store: Optional[bool] = None
    open_store: Optional[bool] = None
    completed_store_exit_survey: Optional[bool] = None
    new_join_info_bar: Optional[bool] = None

    @staticmethod
    def from_dict(obj: Any) -> 'AdminFeatures':
        assert isinstance(obj, dict)
        onboarded_store = from_union([from_bool, from_none], obj.get("onboarded-store"))
        open_store = from_union([from_bool, from_none], obj.get("open-store"))
        completed_store_exit_survey = from_union([from_bool, from_none], obj.get("completed-store-exit-survey"))
        new_join_info_bar = from_union([from_bool, from_none], obj.get("new-join-info-bar"))
        return AdminFeatures(onboarded_store, open_store, completed_store_exit_survey, new_join_info_bar)

    def to_dict(self) -> dict:
        result: dict = {}
        result["onboarded-store"] = from_union([from_bool, from_none], self.onboarded_store)
        result["open-store"] = from_union([from_bool, from_none], self.open_store)
        result["completed-store-exit-survey"] = from_union([from_bool, from_none], self.completed_store_exit_survey)
        result["new-join-info-bar"] = from_union([from_bool, from_none], self.new_join_info_bar)
        return result


class Source(Enum):
    EMAIL_REGISTRATION = "email registration"
    GOOGLE_REGISTRATION = "google registration"
    REGISTRATION = "registration"
    TWITTER_REGISTRATION = "twitter registration"


@dataclass
class Admin:
    id: Optional[str] = None
    valid_email: Optional[bool] = None
    username: Optional[str] = None
    bg_url: Optional[str] = None
    auth0_user_id: Optional[str] = None
    source: Optional[Source] = None
    is_verified: Optional[bool] = None
    timezone: Optional[str] = None
    created_at: Optional[datetime] = None
    email_subscriptions: Optional[AdminEmailSubscriptions] = None
    updated_at: Optional[datetime] = None
    slug: Optional[str] = None
    equipment: Optional[AdminEquipment] = None
    features: Optional[AdminFeatures] = None
    experience_points: Optional[int] = None
    notification_sounds_enabled: Optional[bool] = None
    country_code: Optional[str] = None

    @staticmethod
    def from_dict(obj: Any) -> 'Admin':
        assert isinstance(obj, dict)
        id = from_union([from_str, from_none], obj.get("_id"))
        valid_email = from_union([from_bool, from_none], obj.get("validEmail"))
        username = from_union([from_str, from_none], obj.get("username"))
        bg_url = from_union([from_str, from_none], obj.get("bgUrl"))
        auth0_user_id = from_union([from_str, from_none], obj.get("auth0UserID"))
        source = from_union([Source, from_none], obj.get("source"))
        is_verified = from_union([from_bool, from_none], obj.get("isVerified"))
        timezone = from_union([from_str, from_none], obj.get("timezone"))
        created_at = from_union([from_datetime, from_none], obj.get("createdAt"))
        email_subscriptions = from_union([AdminEmailSubscriptions.from_dict, from_none], obj.get("emailSubscriptions"))
        updated_at = from_union([from_datetime, from_none], obj.get("updatedAt"))
        slug = from_union([from_str, from_none], obj.get("slug"))
        equipment = from_union([AdminEquipment.from_dict, from_none], obj.get("equipment"))
        features = from_union([AdminFeatures.from_dict, from_none], obj.get("features"))
        experience_points = from_union([from_int, from_none], obj.get("experiencePoints"))
        notification_sounds_enabled = from_union([from_bool, from_none], obj.get("notificationSoundsEnabled"))
        country_code = from_union([from_str, from_none], obj.get("countryCode"))
        return Admin(id, valid_email, username, bg_url, auth0_user_id, source, is_verified, timezone, created_at, email_subscriptions, updated_at, slug, equipment, features, experience_points, notification_sounds_enabled, country_code)

    def to_dict(self) -> dict:
        result: dict = {}
        result["_id"] = from_union([from_str, from_none], self.id)
        result["validEmail"] = from_union([from_bool, from_none], self.valid_email)
        result["username"] = from_union([from_str, from_none], self.username)
        result["bgUrl"] = from_union([from_str, from_none], self.bg_url)
        result["auth0UserID"] = from_union([from_str, from_none], self.auth0_user_id)
        result["source"] = from_union([lambda x: to_enum(Source, x), from_none], self.source)
        result["isVerified"] = from_union([from_bool, from_none], self.is_verified)
        result["timezone"] = from_union([from_str, from_none], self.timezone)
        result["createdAt"] = from_union([lambda x: x.isoformat(), from_none], self.created_at)
        result["emailSubscriptions"] = from_union([lambda x: to_class(AdminEmailSubscriptions, x), from_none], self.email_subscriptions)
        result["updatedAt"] = from_union([lambda x: x.isoformat(), from_none], self.updated_at)
        result["slug"] = from_union([from_str, from_none], self.slug)
        result["equipment"] = from_union([lambda x: to_class(AdminEquipment, x), from_none], self.equipment)
        result["features"] = from_union([lambda x: to_class(AdminFeatures, x), from_none], self.features)
        result["experiencePoints"] = from_union([from_int, from_none], self.experience_points)
        result["notificationSoundsEnabled"] = from_union([from_bool, from_none], self.notification_sounds_enabled)
        result["countryCode"] = from_union([from_str, from_none], self.country_code)
        return result


@dataclass
class WelcomeFeatures:
    ladder: Optional[bool] = None

    @staticmethod
    def from_dict(obj: Any) -> 'WelcomeFeatures':
        assert isinstance(obj, dict)
        ladder = from_union([from_bool, from_none], obj.get("ladder"))
        return WelcomeFeatures(ladder)

    def to_dict(self) -> dict:
        result: dict = {}
        result["ladder"] = from_union([from_bool, from_none], self.ladder)
        return result


@dataclass
class Battlenet:
    account_id: Optional[int] = None
    battletag: Optional[str] = None

    @staticmethod
    def from_dict(obj: Any) -> 'Battlenet':
        assert isinstance(obj, dict)
        account_id = from_union([from_none, lambda x: int(from_str(x))], obj.get("accountID"))
        battletag = from_union([from_str, from_none], obj.get("battletag"))
        return Battlenet(account_id, battletag)

    def to_dict(self) -> dict:
        result: dict = {}
        result["accountID"] = from_union([lambda x: from_none((lambda x: is_type(type(None), x))(x)), lambda x: from_str((lambda x: str((lambda x: is_type(int, x))(x)))(x))], self.account_id)
        result["battletag"] = from_union([from_str, from_none], self.battletag)
        return result


@dataclass
class Discord:
    name: Optional[str] = None
    account_id: Optional[str] = None
    username: Optional[str] = None
    discriminator: Optional[str] = None

    @staticmethod
    def from_dict(obj: Any) -> 'Discord':
        assert isinstance(obj, dict)
        name = from_union([from_str, from_none], obj.get("name"))
        account_id = from_union([from_str, from_none], obj.get("accountID"))
        username = from_union([from_str, from_none], obj.get("username"))
        discriminator = from_union([from_str, from_none], obj.get("discriminator"))
        return Discord(name, account_id, username, discriminator)

    def to_dict(self) -> dict:
        result: dict = {}
        result["name"] = from_union([from_str, from_none], self.name)
        result["accountID"] = from_union([from_str, from_none], self.account_id)
        result["username"] = from_union([from_str, from_none], self.username)
        result["discriminator"] = from_union([from_str, from_none], self.discriminator)
        return result


@dataclass
class RANKEDSOLO5X5:
    tier: Optional[str] = None
    division: Optional[str] = None
    last_updated_at: Optional[datetime] = None

    @staticmethod
    def from_dict(obj: Any) -> 'RANKEDSOLO5X5':
        assert isinstance(obj, dict)
        tier = from_union([from_str, from_none], obj.get("tier"))
        division = from_union([from_str, from_none], obj.get("division"))
        last_updated_at = from_union([from_datetime, from_none], obj.get("lastUpdatedAt"))
        return RANKEDSOLO5X5(tier, division, last_updated_at)

    def to_dict(self) -> dict:
        result: dict = {}
        result["tier"] = from_union([from_str, from_none], self.tier)
        result["division"] = from_union([from_str, from_none], self.division)
        result["lastUpdatedAt"] = from_union([lambda x: x.isoformat(), from_none], self.last_updated_at)
        return result


@dataclass
class LeagueOfLegends:
    region: Optional[str] = None
    encrypted_summoner_id: Optional[str] = None
    encrypted_account_id: Optional[str] = None
    summoner_name: Optional[str] = None
    last_updated_at: Optional[datetime] = None
    ranked_solo_5_x5: Optional[RANKEDSOLO5X5] = None

    @staticmethod
    def from_dict(obj: Any) -> 'LeagueOfLegends':
        assert isinstance(obj, dict)
        region = from_union([from_str, from_none], obj.get("region"))
        encrypted_summoner_id = from_union([from_str, from_none], obj.get("encryptedSummonerID"))
        encrypted_account_id = from_union([from_str, from_none], obj.get("encryptedAccountID"))
        summoner_name = from_union([from_str, from_none], obj.get("summonerName"))
        last_updated_at = from_union([from_datetime, from_none], obj.get("lastUpdatedAt"))
        ranked_solo_5_x5 = from_union([RANKEDSOLO5X5.from_dict, from_none], obj.get("RANKED_SOLO_5x5"))
        return LeagueOfLegends(region, encrypted_summoner_id, encrypted_account_id, summoner_name, last_updated_at, ranked_solo_5_x5)

    def to_dict(self) -> dict:
        result: dict = {}
        result["region"] = from_union([from_str, from_none], self.region)
        result["encryptedSummonerID"] = from_union([from_str, from_none], self.encrypted_summoner_id)
        result["encryptedAccountID"] = from_union([from_str, from_none], self.encrypted_account_id)
        result["summonerName"] = from_union([from_str, from_none], self.summoner_name)
        result["lastUpdatedAt"] = from_union([lambda x: x.isoformat(), from_none], self.last_updated_at)
        result["RANKED_SOLO_5x5"] = from_union([lambda x: to_class(RANKEDSOLO5X5, x), from_none], self.ranked_solo_5_x5)
        return result


@dataclass
class Twitter:
    screen_name: Optional[str] = None
    account_id: Optional[str] = None

    @staticmethod
    def from_dict(obj: Any) -> 'Twitter':
        assert isinstance(obj, dict)
        screen_name = from_union([from_str, from_none], obj.get("screen_name"))
        account_id = from_union([from_str, from_none], obj.get("accountID"))
        return Twitter(screen_name, account_id)

    def to_dict(self) -> dict:
        result: dict = {}
        result["screen_name"] = from_union([from_str, from_none], self.screen_name)
        result["accountID"] = from_union([from_str, from_none], self.account_id)
        return result


@dataclass
class Accounts:
    twitch: Optional[str] = None
    twitter: Optional[Twitter] = None
    discord: Optional[Discord] = None
    battlenet: Optional[Battlenet] = None
    league_of_legends: Optional[LeagueOfLegends] = None

    @staticmethod
    def from_dict(obj: Any) -> 'Accounts':
        assert isinstance(obj, dict)
        twitch = from_union([from_str, from_none], obj.get("twitch"))
        twitter = from_union([Twitter.from_dict, from_none], obj.get("twitter"))
        discord = from_union([Discord.from_dict, from_none], obj.get("discord"))
        battlenet = from_union([Battlenet.from_dict, from_none], obj.get("battlenet"))
        league_of_legends = from_union([LeagueOfLegends.from_dict, from_none], obj.get("league-of-legends"))
        return Accounts(twitch, twitter, discord, battlenet, league_of_legends)

    def to_dict(self) -> dict:
        result: dict = {}
        result["twitch"] = from_union([from_str, from_none], self.twitch)
        result["twitter"] = from_union([lambda x: to_class(Twitter, x), from_none], self.twitter)
        result["discord"] = from_union([lambda x: to_class(Discord, x), from_none], self.discord)
        result["battlenet"] = from_union([lambda x: to_class(Battlenet, x), from_none], self.battlenet)
        result["league-of-legends"] = from_union([lambda x: to_class(LeagueOfLegends, x), from_none], self.league_of_legends)
        return result


@dataclass
class ModeratorEmailSubscriptions:
    weekly_recommendations: Optional[bool] = None
    check_ins: Optional[bool] = None

    @staticmethod
    def from_dict(obj: Any) -> 'ModeratorEmailSubscriptions':
        assert isinstance(obj, dict)
        weekly_recommendations = from_union([from_bool, from_none], obj.get("weeklyRecommendations"))
        check_ins = from_union([from_bool, from_none], obj.get("checkIns"))
        return ModeratorEmailSubscriptions(weekly_recommendations, check_ins)

    def to_dict(self) -> dict:
        result: dict = {}
        result["weeklyRecommendations"] = from_union([from_bool, from_none], self.weekly_recommendations)
        result["checkIns"] = from_union([from_bool, from_none], self.check_ins)
        return result


@dataclass
class ModeratorEquipment:
    face: Optional[str] = None
    head: Optional[str] = None
    body: Optional[str] = None

    @staticmethod
    def from_dict(obj: Any) -> 'ModeratorEquipment':
        assert isinstance(obj, dict)
        face = from_union([from_str, from_none], obj.get("face"))
        head = from_union([from_str, from_none], obj.get("head"))
        body = from_union([from_str, from_none], obj.get("body"))
        return ModeratorEquipment(face, head, body)

    def to_dict(self) -> dict:
        result: dict = {}
        result["face"] = from_union([from_str, from_none], self.face)
        result["head"] = from_union([from_str, from_none], self.head)
        result["body"] = from_union([from_str, from_none], self.body)
        return result


@dataclass
class Moderator:
    id: Optional[str] = None
    valid_email: Optional[bool] = None
    username: Optional[str] = None
    bg_url: Optional[str] = None
    auth0_user_id: Optional[str] = None
    source: Optional[Source] = None
    is_verified: Optional[bool] = None
    timezone: Optional[str] = None
    created_at: Optional[datetime] = None
    email_subscriptions: Optional[ModeratorEmailSubscriptions] = None
    updated_at: Optional[datetime] = None
    slug: Optional[str] = None
    equipment: Optional[ModeratorEquipment] = None
    features: Optional[AdminFeatures] = None
    experience_points: Optional[int] = None
    notification_sounds_enabled: Optional[bool] = None
    country_code: Optional[str] = None
    accounts: Optional[Accounts] = None
    normalized_username: Optional[str] = None
    has_seen_vgl_ad_at: Optional[datetime] = None
    has_seen_ncsa_ad_at: Optional[datetime] = None
    avatar_url: Optional[str] = None

    @staticmethod
    def from_dict(obj: Any) -> 'Moderator':
        assert isinstance(obj, dict)
        id = from_union([from_str, from_none], obj.get("_id"))
        valid_email = from_union([from_bool, from_none], obj.get("validEmail"))
        username = from_union([from_str, from_none], obj.get("username"))
        bg_url = from_union([from_str, from_none], obj.get("bgUrl"))
        auth0_user_id = from_union([from_str, from_none], obj.get("auth0UserID"))
        source = from_union([Source, from_none], obj.get("source"))
        is_verified = from_union([from_bool, from_none], obj.get("isVerified"))
        timezone = from_union([from_str, from_none], obj.get("timezone"))
        created_at = from_union([from_datetime, from_none], obj.get("createdAt"))
        email_subscriptions = from_union([ModeratorEmailSubscriptions.from_dict, from_none], obj.get("emailSubscriptions"))
        updated_at = from_union([from_datetime, from_none], obj.get("updatedAt"))
        slug = from_union([from_str, from_none], obj.get("slug"))
        equipment = from_union([ModeratorEquipment.from_dict, from_none], obj.get("equipment"))
        features = from_union([AdminFeatures.from_dict, from_none], obj.get("features"))
        experience_points = from_union([from_int, from_none], obj.get("experiencePoints"))
        notification_sounds_enabled = from_union([from_bool, from_none], obj.get("notificationSoundsEnabled"))
        country_code = from_union([from_str, from_none], obj.get("countryCode"))
        accounts = from_union([Accounts.from_dict, from_none], obj.get("accounts"))
        normalized_username = from_union([from_str, from_none], obj.get("normalizedUsername"))
        has_seen_vgl_ad_at = from_union([from_datetime, from_none], obj.get("hasSeenVGLAdAt"))
        has_seen_ncsa_ad_at = from_union([from_datetime, from_none], obj.get("hasSeenNCSAAdAt"))
        avatar_url = from_union([from_str, from_none], obj.get("avatarUrl"))
        return Moderator(id, valid_email, username, bg_url, auth0_user_id, source, is_verified, timezone, created_at, email_subscriptions, updated_at, slug, equipment, features, experience_points, notification_sounds_enabled, country_code, accounts, normalized_username, has_seen_vgl_ad_at, has_seen_ncsa_ad_at, avatar_url)

    def to_dict(self) -> dict:
        result: dict = {}
        result["_id"] = from_union([from_str, from_none], self.id)
        result["validEmail"] = from_union([from_bool, from_none], self.valid_email)
        result["username"] = from_union([from_str, from_none], self.username)
        result["bgUrl"] = from_union([from_str, from_none], self.bg_url)
        result["auth0UserID"] = from_union([from_str, from_none], self.auth0_user_id)
        result["source"] = from_union([lambda x: to_enum(Source, x), from_none], self.source)
        result["isVerified"] = from_union([from_bool, from_none], self.is_verified)
        result["timezone"] = from_union([from_str, from_none], self.timezone)
        result["createdAt"] = from_union([lambda x: x.isoformat(), from_none], self.created_at)
        result["emailSubscriptions"] = from_union([lambda x: to_class(ModeratorEmailSubscriptions, x), from_none], self.email_subscriptions)
        result["updatedAt"] = from_union([lambda x: x.isoformat(), from_none], self.updated_at)
        result["slug"] = from_union([from_str, from_none], self.slug)
        result["equipment"] = from_union([lambda x: to_class(ModeratorEquipment, x), from_none], self.equipment)
        result["features"] = from_union([lambda x: to_class(AdminFeatures, x), from_none], self.features)
        result["experiencePoints"] = from_union([from_int, from_none], self.experience_points)
        result["notificationSoundsEnabled"] = from_union([from_bool, from_none], self.notification_sounds_enabled)
        result["countryCode"] = from_union([from_str, from_none], self.country_code)
        result["accounts"] = from_union([lambda x: to_class(Accounts, x), from_none], self.accounts)
        result["normalizedUsername"] = from_union([from_str, from_none], self.normalized_username)
        result["hasSeenVGLAdAt"] = from_union([lambda x: x.isoformat(), from_none], self.has_seen_vgl_ad_at)
        result["hasSeenNCSAAdAt"] = from_union([lambda x: x.isoformat(), from_none], self.has_seen_ncsa_ad_at)
        result["avatarUrl"] = from_union([from_str, from_none], self.avatar_url)
        return result


@dataclass
class Owner:
    id: Optional[str] = None
    valid_email: Optional[bool] = None
    username: Optional[str] = None
    bg_url: Optional[str] = None
    auth0_user_id: Optional[str] = None
    source: Optional[Source] = None
    is_verified: Optional[bool] = None
    timezone: Optional[str] = None
    created_at: Optional[datetime] = None
    email_subscriptions: Optional[AdminEmailSubscriptions] = None
    updated_at: Optional[datetime] = None
    slug: Optional[str] = None
    equipment: Optional[ModeratorEquipment] = None
    features: Optional[AdminFeatures] = None
    experience_points: Optional[int] = None
    notification_sounds_enabled: Optional[bool] = None
    country_code: Optional[str] = None

    @staticmethod
    def from_dict(obj: Any) -> 'Owner':
        assert isinstance(obj, dict)
        id = from_union([from_str, from_none], obj.get("_id"))
        valid_email = from_union([from_bool, from_none], obj.get("validEmail"))
        username = from_union([from_str, from_none], obj.get("username"))
        bg_url = from_union([from_str, from_none], obj.get("bgUrl"))
        auth0_user_id = from_union([from_str, from_none], obj.get("auth0UserID"))
        source = from_union([Source, from_none], obj.get("source"))
        is_verified = from_union([from_bool, from_none], obj.get("isVerified"))
        timezone = from_union([from_str, from_none], obj.get("timezone"))
        created_at = from_union([from_datetime, from_none], obj.get("createdAt"))
        email_subscriptions = from_union([AdminEmailSubscriptions.from_dict, from_none], obj.get("emailSubscriptions"))
        updated_at = from_union([from_datetime, from_none], obj.get("updatedAt"))
        slug = from_union([from_str, from_none], obj.get("slug"))
        equipment = from_union([ModeratorEquipment.from_dict, from_none], obj.get("equipment"))
        features = from_union([AdminFeatures.from_dict, from_none], obj.get("features"))
        experience_points = from_union([from_int, from_none], obj.get("experiencePoints"))
        notification_sounds_enabled = from_union([from_bool, from_none], obj.get("notificationSoundsEnabled"))
        country_code = from_union([from_str, from_none], obj.get("countryCode"))
        return Owner(id, valid_email, username, bg_url, auth0_user_id, source, is_verified, timezone, created_at, email_subscriptions, updated_at, slug, equipment, features, experience_points, notification_sounds_enabled, country_code)

    def to_dict(self) -> dict:
        result: dict = {}
        result["_id"] = from_union([from_str, from_none], self.id)
        result["validEmail"] = from_union([from_bool, from_none], self.valid_email)
        result["username"] = from_union([from_str, from_none], self.username)
        result["bgUrl"] = from_union([from_str, from_none], self.bg_url)
        result["auth0UserID"] = from_union([from_str, from_none], self.auth0_user_id)
        result["source"] = from_union([lambda x: to_enum(Source, x), from_none], self.source)
        result["isVerified"] = from_union([from_bool, from_none], self.is_verified)
        result["timezone"] = from_union([from_str, from_none], self.timezone)
        result["createdAt"] = from_union([lambda x: x.isoformat(), from_none], self.created_at)
        result["emailSubscriptions"] = from_union([lambda x: to_class(AdminEmailSubscriptions, x), from_none], self.email_subscriptions)
        result["updatedAt"] = from_union([lambda x: x.isoformat(), from_none], self.updated_at)
        result["slug"] = from_union([from_str, from_none], self.slug)
        result["equipment"] = from_union([lambda x: to_class(ModeratorEquipment, x), from_none], self.equipment)
        result["features"] = from_union([lambda x: to_class(AdminFeatures, x), from_none], self.features)
        result["experiencePoints"] = from_union([from_int, from_none], self.experience_points)
        result["notificationSoundsEnabled"] = from_union([from_bool, from_none], self.notification_sounds_enabled)
        result["countryCode"] = from_union([from_str, from_none], self.country_code)
        return result


@dataclass
class Social:
    website_url: Optional[str] = None
    twitchtv_url: Optional[str] = None
    youtube_url: Optional[str] = None
    twitter_url: Optional[str] = None
    discord_url: Optional[str] = None
    facebook_url: Optional[str] = None

    @staticmethod
    def from_dict(obj: Any) -> 'Social':
        assert isinstance(obj, dict)
        website_url = from_union([from_str, from_none], obj.get("websiteUrl"))
        twitchtv_url = from_union([from_str, from_none], obj.get("twitchtvUrl"))
        youtube_url = from_union([from_str, from_none], obj.get("youtubeUrl"))
        twitter_url = from_union([from_str, from_none], obj.get("twitterUrl"))
        discord_url = from_union([from_str, from_none], obj.get("discordUrl"))
        facebook_url = from_union([from_str, from_none], obj.get("facebookUrl"))
        return Social(website_url, twitchtv_url, youtube_url, twitter_url, discord_url, facebook_url)

    def to_dict(self) -> dict:
        result: dict = {}
        result["websiteUrl"] = from_union([from_str, from_none], self.website_url)
        result["twitchtvUrl"] = from_union([from_str, from_none], self.twitchtv_url)
        result["youtubeUrl"] = from_union([from_str, from_none], self.youtube_url)
        result["twitterUrl"] = from_union([from_str, from_none], self.twitter_url)
        result["discordUrl"] = from_union([from_str, from_none], self.discord_url)
        result["facebookUrl"] = from_union([from_str, from_none], self.facebook_url)
        return result


@dataclass
class Organization:
    id: Optional[str] = None
    name: Optional[str] = None
    owner_id: Optional[str] = None
    admin_i_ds: Optional[List[str]] = None
    moderator_i_ds: Optional[List[str]] = None
    runner_i_ds: Optional[List[Any]] = None
    features: Optional[WelcomeFeatures] = None
    created_at: Optional[datetime] = None
    service_fee_percent: Optional[int] = None
    slug: Optional[str] = None
    social: Optional[Social] = None
    logo_url: Optional[str] = None
    short_description: Optional[str] = None
    followers: Optional[int] = None
    q_score: Optional[float] = None
    banner_url: Optional[str] = None
    type: Optional[str] = None
    cloud_search_document_hash: Optional[str] = None
    cloud_search_document_last_generated: Optional[datetime] = None
    visible_tournaments: Optional[int] = None

    @staticmethod
    def from_dict(obj: Any) -> 'WelcomeElement':
        assert isinstance(obj, dict)
        id = from_union([from_str, from_none], obj.get("_id"))
        name = from_union([from_str, from_none], obj.get("name"))
        owner_id = from_union([from_str, from_none], obj.get("ownerID"))
        admin_i_ds = from_union([lambda x: from_list(from_str, x), from_none], obj.get("adminIDs"))
        moderator_i_ds = from_union([lambda x: from_list(from_str, x), from_none], obj.get("moderatorIDs"))
        runner_i_ds = from_union([lambda x: from_list(lambda x: x, x), from_none], obj.get("runnerIDs"))
        features = from_union([WelcomeFeatures.from_dict, from_none], obj.get("features"))
        created_at = from_union([from_datetime, from_none], obj.get("createdAt"))
        service_fee_percent = from_union([from_int, from_none], obj.get("serviceFeePercent"))
        slug = from_union([from_str, from_none], obj.get("slug"))
        social = from_union([Social.from_dict, from_none], obj.get("social"))
        logo_url = from_union([from_str, from_none], obj.get("logoUrl"))
        short_description = from_union([from_str, from_none], obj.get("shortDescription"))
        followers = from_union([from_int, from_none], obj.get("followers"))
        q_score = from_union([from_float, from_none], obj.get("qScore"))
        banner_url = from_union([from_str, from_none], obj.get("bannerUrl"))
        type = from_union([from_str, from_none], obj.get("type"))
        cloud_search_document_hash = from_union([from_str, from_none], obj.get("cloudSearchDocumentHash"))
        cloud_search_document_last_generated = from_union([from_datetime, from_none], obj.get("cloudSearchDocumentLastGenerated"))
        visible_tournaments = from_union([from_int, from_none], obj.get("visibleTournaments"))
        return Organization(id, name, owner_id, admin_i_ds, moderator_i_ds, runner_i_ds, features, created_at, service_fee_percent, slug, social, logo_url, short_description, followers, q_score, banner_url, type, cloud_search_document_hash, cloud_search_document_last_generated, visible_tournaments)

    def to_dict(self) -> dict:
        result: dict = {}
        result["_id"] = from_union([from_str, from_none], self.id)
        result["name"] = from_union([from_str, from_none], self.name)
        result["ownerID"] = from_union([from_str, from_none], self.owner_id)
        result["adminIDs"] = from_union([lambda x: from_list(from_str, x), from_none], self.admin_i_ds)
        result["moderatorIDs"] = from_union([lambda x: from_list(from_str, x), from_none], self.moderator_i_ds)
        result["runnerIDs"] = from_union([lambda x: from_list(lambda x: x, x), from_none], self.runner_i_ds)
        result["features"] = from_union([lambda x: to_class(WelcomeFeatures, x), from_none], self.features)
        result["createdAt"] = from_union([lambda x: x.isoformat(), from_none], self.created_at)
        result["serviceFeePercent"] = from_union([from_int, from_none], self.service_fee_percent)
        result["slug"] = from_union([from_str, from_none], self.slug)
        result["social"] = from_union([lambda x: to_class(Social, x), from_none], self.social)
        result["logoUrl"] = from_union([from_str, from_none], self.logo_url)
        result["shortDescription"] = from_union([from_str, from_none], self.short_description)
        result["followers"] = from_union([from_int, from_none], self.followers)
        result["qScore"] = from_union([to_float, from_none], self.q_score)
        result["bannerUrl"] = from_union([from_str, from_none], self.banner_url)
        result["type"] = from_union([from_str, from_none], self.type)
        result["cloudSearchDocumentHash"] = from_union([from_str, from_none], self.cloud_search_document_hash)
        result["cloudSearchDocumentLastGenerated"] = from_union([lambda x: x.isoformat(), from_none], self.cloud_search_document_last_generated)
        result["visibleTournaments"] = from_union([from_int, from_none], self.visible_tournaments)
        return result


def organization_from_dict(s: Any) -> List[Organization]:
    return from_list(Organization.from_dict, s)
